#############################################################################
#
# tclsvc work by matt newman (www.sensus.org/tcl)
# setup tool by michael kraus (mkextensions.sourceforge.net)
#
# script to install (and uninstall) tclsvc as a service. essentially it
# performs the steps described in matt's tclsvc.html. in addition, it
# displays currently installed tclsvc-based services (that is, those, where
# where the registry value 'TclScript' exists), and lets you uninstall such
# a service. since this slightly enhanced tclsvc version from me supports
# the windows event log, it also creates an additional registry key.
#
# note that error handling is not quite reliable, because the calls
# "tclsvc -install" and "tclsvc -remove" do not return with error
# even when they fail (e.g. due to a duplicate service display name).
# this is a windows problem, not a tclsvc problem. check the messages.
#
# 2004-03-03: initial version
#
#############################################################################

# these registry keys will be accessed

set sServicesKey {HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services}
set sEventlogKey {HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet\Services\Eventlog\Application}

# where are we, where do we go, what version am i?

set sSourcePath  [file dirname [file normalize $argv0]]
set sTargetPath  [file dirname [info nameofexecutable]]
set sServiceExe  tclsvc[string map {. {}} [info tclversion]].exe

# parameter default values

set sAction      Install
set sShortname   tclsvc84
set sLongname    "Tcl Service"
set sTclscript   ""

# lists currently installed services based on tclsvc

proc showServices {} {
  package require registry

  # find services where the value TclScript exists

  foreach sShortname [registry key $::sServicesKey] {
    if { ! [catch { registry get $::sServicesKey\\$sShortname TclScript } sTclscript] } {
      set sLongname "(undefined)"
      catch { set sLongname [registry get $::sServicesKey\\$sShortname DisplayName] }
      lappend lServices $sShortname $sLongname $sTclscript
    }
  }

  # list them, if we found some

  if { ! [info exists lServices] } {
    log "No installed Tcl services found"
  } else {
    log "Installed Tcl services:"
    foreach {sShortname sLongname sTclscript} $lServices {
      log "  Short Name: '$sShortname'; Long Name: '$sLongname'; Tcl Script: '$sTclscript'"
    }
    log "End of list"
  }
}

# installs a service with tslsvc

proc installService { sShortname sLongname sTclscript } {
  package require registry

  set sMessage "The service $sShortname (\"$sLongname\")\nwill now be installed."
  if { [tk_messageBox -message $sMessage -type okcancel -default cancel -icon warning] == "cancel" } return

  # read the log messages below. we follow matt's installation steps.

  set sSourceExe  $::sSourcePath/$::sServiceExe
  set sTargetExe  $::sTargetPath/$sShortname.exe
  set sServiceKey $::sServicesKey\\$sShortname
  set sSvcEvntKey $::sEventlogKey\\$sShortname

  log "Installing service $sShortname..."
  if { [catch {
    log "  Copying [file tail $sSourceExe] to $sTargetExe"
    file copy -force $sSourceExe $sTargetExe
    log "  Installing service $sShortname with display name '$sLongname'"
    set sResult [exec $sTargetExe -install $sLongname]
    log "  Result: $sResult"
    log "  Creating registry value $sServiceKey\\TclScript"
    registry set $sServiceKey TclScript $sTclscript
    log "  Creating registry key $sSvcEvntKey"
    registry set $sSvcEvntKey EventMessageFile [file nativename $::sTargetPath/$sShortname.exe]
    registry set $sSvcEvntKey TypesSupported   7 dword
  } sError] } {
    log "Installation failed ($sError)"
  } else {
    log "Installation finished"
  }
}

# uninstalls a service

proc uninstallService { sShortname } {
  package require registry

  set sMessage "The service $sShortname\nwill now be uninstalled."
  if { [tk_messageBox -message $sMessage -type okcancel -default cancel -icon warning] == "cancel" } return

  # read log messages. this is simply the inverse of the installation steps.

  set sTargetExe  $::sTargetPath/$sShortname.exe
  set sServiceKey $::sServicesKey\\$sShortname
  set sSvcEvntKey $::sEventlogKey\\$sShortname

  log "Deinstalling service $sShortname..."
  if { [catch {
    log "  Removing service $sShortname"
    set sResult [exec $sTargetExe -remove]
    log "  Result: $sResult"
    log "  Removing executable $sTargetExe"
    file delete -force $sTargetExe
    log "  Removing registry key $sSvcEvntKey"
    registry delete $sSvcEvntKey
    log "  Removing registry key $sServiceKey"
    registry delete $sServiceKey
  } sError] } {
    log "Deinstallation failed ($sError)"
    log "Either the service does not exist, or its configuration is somehow corrupted."
    log "You can deinstall and clean up manually:"
    log "- Remove registry key $sServiceKey."
    log "- Remove registry key $sSvcEvntKey."
    log "- If possible, call '$sTargetExe -remove', then delete the file."
    log "- If the services list still shows a zombie, reboot the system."
  } else {
    log "Deinstallation finished"
  }
}

# starts any service

proc startService { sShortname } {
  set sMessage "Service $sShortname will now be started."
  if { [tk_messageBox -message $sMessage -type okcancel -default cancel -icon warning] == "cancel" } return

  # unfortunately, "net" does not pass its stdout to exec.

  log "Starting service $sShortname..."
  if { [catch { exec net start $sShortname }] } {
    log "Service $sShortname could not be started"
  } else {
    log "Service $sShortname started"
  }
}

# stops any service

proc stopService { sShortname } {
  set sMessage "Service $sShortname will now be stopped."
  if { [tk_messageBox -message $sMessage -type okcancel -default cancel -icon warning] == "cancel" } return

  log "Stopping service $sShortname..."
  if { [catch { exec net stop $sShortname }] } {
    log "Service $sShortname could not be stopped"
  } else {
    log "Service $sShortname stopped"
  }
}

# dispatcher for action selected by the user

proc doAction { sAction } {
  switch $sAction {
    Install   { installService   $::sShortname $::sLongname $::sTclscript; showServices }
    Uninstall { uninstallService $::sShortname; showServices }
    Start     { startService     $::sShortname }
    Stop      { stopService      $::sShortname }
    default   { error "invalid action $sAction: must be Install, Uninstall, Start or Stop" }
  }
}

# default implementation of log proc (user interface may redefine)

proc log { sText } {
  puts $sText
}

# user interfaces start here. currently only tk-based is supported.

if { ! [catch { package present Tk }] } {
  wm title . "Tclsvc Setup"

  pack [label .logo -text "T\nc\nl\ns\nv\nc" -border 4 -relief ridge -bg orange -padx 15 -font {Courier 14 bold italic}] -side left -fill y -padx 2 -pady 2
  pack [frame .dist -width 3] -side left

  pack [labelframe  .what -text " Action "] -fill x
  grid [radiobutton .what.inst -text "Install Tcl Service"   -variable sAction -value Install   -command setentries -padx 4] -column 0 -row 0 -sticky w
  grid [radiobutton .what.unin -text "Uninstall Tcl Service" -variable sAction -value Uninstall -command setentries -padx 4] -column 0 -row 1 -sticky w
  grid [radiobutton .what.star -text "Start Service"         -variable sAction -value Start     -command setentries -padx 4] -column 1 -row 0 -sticky w
  grid [radiobutton .what.stop -text "Stop Service"          -variable sAction -value Stop      -command setentries -padx 4] -column 1 -row 1 -sticky w

  pack [labelframe  .sett -text " Settings "] -fill x
  grid [label       .sett.cap1 -text "Short name:"     ] -column 0 -row 0 -sticky w
  grid [entry       .sett.snam -textvariable sShortname] -column 1 -row 0 -sticky we -padx 4 -pady 1
  grid [label       .sett.cap2 -text "Long name:"      ] -column 0 -row 1 -sticky w
  grid [entry       .sett.lnam -textvariable sLongname ] -column 1 -row 1 -sticky we -padx 4 -pady 1
  grid [label       .sett.cap3 -text "Tcl Script:"     ] -column 0 -row 2 -sticky w
  grid [entry       .sett.tcls -textvariable sTclscript] -column 1 -row 2 -sticky we -padx 4 -pady 1
  grid [frame       .sett.dist -height 4               ] -column 0 -row 3
  grid columnconfigure .sett 1 -weight 1

  pack [labelframe  .logb -text " Messages "] -side top -fill both -expand 1
  pack [frame       .logb.dist -border 4 -relief flat] -fill both -expand 1
  pack [scrollbar   .logb.dist.scrv -command ".logb.dist.list yview"] -side right -fill y
  pack [listbox     .logb.dist.list -yscroll ".logb.dist.scrv set" -height 6 -selectborder 0 -border 1] -fill both -expand 1

  pack [button      .actn -textvariable sAction -command {doAction $::sAction}] -fill x -padx 4 -pady 4

  proc setentries {} {
    if { $::sAction == "Install" } {
       set sState normal
    } else {
       set sState disabled
    }

    .sett.lnam configure -state $sState
    .sett.tcls configure -state $sState
  }

  proc log { sText } {
    .logb.dist.list insert end $sText
    .logb.dist.list see end
    update idletasks
  }

  after idle {
    after 0 showServices
  }
} elseif { ! [llength $argv] } {
  puts "Sorry, command line interface not implemented."
  exit 1
}
